module.exports = function parse(input) {
  var orig = input
    , thunks = []
    , vstack = []
    , rightmostPos = 0
    , pos = 0
    , begin = 0
    , end = 0
    , thunkpos
    , mark
    , text
    , tmp
    , ts
    , $;

  cookie = {};

  return _cookie(), done();

  function done() {
    var len = thunks.length
      , thunk;
    for (var i = 0; i < len; ++i) {
      thunk = thunks[i];
      $ = undefined;
      thunk(thunk.begin, thunk.end);
      if (undefined !== $) vstack.push($);
    }
    if (input.length) error();
    return $;
  }

  function inspect(str) {
    return str
      .replace(/\\/g, '\\\\')
      .replace(/\r/g, '\\r')
      .replace(/\n/g, '\\n')
      .replace(/"/g, '\\"');
  }

  function state() {
    rightmostPos = pos;
    return {
        thunkpos: thunks.length
      , input: input
      , pos: pos
    };
  }

  function restore(state) {
    thunks.length = state.thunkpos;
    input = state.input;
    pos = state.pos;
    return 1;
  }

  function thunk(fn) {
    fn.begin = begin;
    fn.end = end;
    thunks.push(fn);
    return 1;
  }

  function anyChar() {
    if (input.length) {
      ++pos;
      input = input.substr(1);
      return 1;
    }
  }

  function char(c) {
    if (c == input[0]) {
      ++pos;
      input = input.substr(1);
      return 1;
    }
  }

  function string(str) {
    if (0 == input.indexOf(str)) {
      var len = str.length;
      pos += len;
      input = input.substr(len);
      return 1;
    }
  }

  function pattern(regexp) {
    var captures;
    if (captures = regexp.exec(input)) {
      var len = captures[0].length;
      pos += len;
      input = input.substr(len);
      return 1;
    }
  }

  function errorPosition() {
    var lineno = 1;
    for (var i = 0; i < pos; ++i) {
      switch (orig[i]) {
        case '\n': ++lineno; break;
      }
    }
    return { line: lineno };
  }

  function error() {
    pos = rightmostPos;
    var prefix = 'parse error on line ' + errorPosition().line + ' near '
      , prefixlen = 'Error: '.length + prefix.length + 1
      , context = 10
      , start = Math.max(pos - context, 0)
      , len = orig.length
      , end = Math.min(pos + context, len)
      , mark = pos - start
      , tilde = Array(end - pos).join('~')
      , ellipsis = end == len ? '' : '...'
      , _context = orig.slice(start, end) + ellipsis
      , context = inspect(_context)
      , head = _context.slice(0, mark)
      , diff = inspect(head).length - head.length
      , mark = Array(prefixlen + 1 + diff + mark).join(' ') + '^' + tilde;
    throw new Error(prefix + '"' + context + '"' + '\n' + mark);
  }
    
  function _cookie(){
    var s = state();
    if ((string("Cookie")
        && __()
        && char(":")
        && (function(){
        if (!(function(){
          var s = state();
          return (__()
            && _id()
            && __()
            && char("=")
            && __()
            && _value()
            && __()
            && (char(";") || 1)
            && thunk(function(){
            var val = vstack.pop()
              , name = vstack.pop();
            cookie[name] = val
          }));
        })()) return;
        while ((function(){
          var s = state();
          return (__()
            && _id()
            && __()
            && char("=")
            && __()
            && _value()
            && __()
            && (char(";") || 1)
              && thunk(function(){
            var val = vstack.pop()
              , name = vstack.pop();
            cookie[name] = val
          }));
        })()) ;
        return 1;
      })()
        && thunk(function(){
        $ = cookie
      }))) return 1;
    restore(s);
  }
  
  function _value(){
    var s = state();
    if ((_string()
      || (restore(s)
       && _id()))) return 1;
    restore(s);
  }
  
  function _string(){
    var s = state();
    if ((((begin=pos, 1)
      && pattern(/^"([^"]+)"/)
      && (end=pos, 1)
      && thunk(function(begin, end){ text = orig.slice(begin, end); }))
        && thunk(function(){
        $ = text
      }))) return 1;
    restore(s);
  }
  
  function _id(){
    var s = state();
    if ((((begin=pos, 1)
      && pattern(/^[\w]+/)
      && (end=pos, 1)
      && thunk(function(begin, end){ text = orig.slice(begin, end); }))
        && thunk(function(){
        $ = text
      }))) return 1;
    restore(s);
  }
  
  function __(){
    var s = state();
    if (pattern(/^[ \t]*/)) return 1;
    restore(s);
  }
  
}