
/*!
 * descent - lexer
 * Copyright (c) 2011 LearnBoost <tj@learnboost.com>
 * MIT Licensed
 */

/**
 * Perform lexical analysis on the given `input` string,
 * returning a sequence of tokens.
 *
 * @param {String} input
 * @return {Array}
 * @api private
 */

module.exports = function(input) {
  var toks = []
    , lineno = 1;

  // TODO: self-hosted

  /**
   * Scan `str` with optional `name` and/or callback `fn`.
   */

  function scanString(str, name, fn) {
    if (0 == input.indexOf(str)) {
      name = name || str;
      input = input.substr(str.length);
      if (fn) fn(str, name);
      var tok = [name];
      return tok;
    }
  }

  /**
   * Scan `regexp` with optional `name`.
   */

  function scanPattern(regexp, name, ignore) {
    var captures;
    if (captures = regexp.exec(input)) {
      input = input.substr(captures[0].length);
      if (ignore) return next();
      captures[0] = name;
      delete captures.index;
      delete captures.input;
      return captures;
    }
  }

  /**
   * Scan action by mapping braces.
   */

  function scanAction() {
    var buf = ''
      , braces = 1;
    if ('{' == input[0]) {
      for (var i = 1, len = input.length; i < len; ++i) {
        switch (input[i]) {
          case '{': ++braces;
          default: buf += input[i]; break;
          case '}':
            if (!--braces) {
              input = input.substr(++i);
              return ['action', buf];
            } else {
              buf += '}';
            }
        }
      }
    }
  }

  /**
   * throw syntax error.
   */

  function error() {
    var msg = 'SyntaxError at "' + input.slice(0, 10) + '" on line ' + lineno;
    throw new Error(msg);
  }

  /**
   * Return the next token.
   */

  function next() {
    return scanString('=')
      || scanString(':')
      || scanString(';')
      || scanString('|')
      || scanString('*')
      || scanString('+')
      || scanString('?')
      || scanString('(')
      || scanString(')')
      || scanString('<')
      || scanString('>')
      || scanString('&')
      || scanString('.')
      || scanString('!')
      || scanString('\n', 'newline', function(){ ++lineno; })
      || scanPattern(/^[ \t]+/, 'space', true)
      || scanPattern(/^(\[[^\]]+\](?:[+*]|\{\d+\})?)/, 'regexp')
      || scanPattern(/^\/([^\/]+)\/(i?)/, 'regexp')
      || scanPattern(/^"([^"]*)"/, 'string')
      || scanPattern(/^'([^']*)'/, 'string')
      || scanPattern(/^([-\w]+)/, 'rule')
      || scanPattern(/^\/\/([^\n]*)/, 'comment', true)
      || scanAction()
      || error();
  }

  /**
   * Scan entire string.
   */

  while (input.length) {
    toks.push(next());
  }

  return toks;
};